package cz.vydrzel.dpczu.service;

import com.auth0.jwt.JWT;
import com.auth0.jwt.algorithms.Algorithm;
import com.auth0.jwt.interfaces.DecodedJWT;
import cz.vydrzel.dpczu.api.demoapp.model.Profile;
import cz.vydrzel.dpczu.dao.UserDao;
import cz.vydrzel.dpczu.dao.model.UserEntity;
import jakarta.annotation.PostConstruct;
import jakarta.transaction.Transactional;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;

import java.security.Security;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPublicKey;
import java.util.List;
import java.util.stream.StreamSupport;

import static org.springframework.http.HttpStatus.FORBIDDEN;
import static org.springframework.http.HttpStatus.UNAUTHORIZED;

@Slf4j
@Service
@Transactional
@RequiredArgsConstructor
public class DemoAppService {

    private final UserDao userDao;

    private RSAPublicKey jwtPublicKey;

    @PostConstruct
    @SneakyThrows
    public void init() {
        Security.addProvider(new BouncyCastleProvider());

        CertificateFactory certificateFactory = CertificateFactory.getInstance("X.509");
        X509Certificate x509Certificate = (X509Certificate) certificateFactory.generateCertificate(getClass().getResourceAsStream("/certificates/jwt-certificate.pem"));
        jwtPublicKey = (RSAPublicKey) x509Certificate.getPublicKey();
    }

    public Profile getMyProfile(String authorizationHeader) {
        UserEntity userEntity = verifyAccessToken(authorizationHeader, "profile");

        return Profile.builder()
                .email(userEntity.getLogin())
                .role(userEntity.getRole())
                .firstname(userEntity.getFirstname())
                .lastname(userEntity.getLastname())
                .build();
    }

    public List<Profile> getAllProfiles(String authorizationHeader) {
        verifyAccessToken(authorizationHeader, "users_list");

        return StreamSupport.stream(userDao.findAll().spliterator(), false)
                .map(user -> Profile.builder()
                        .email(user.getLogin())
                        .role(user.getRole())
                        .firstname(user.getFirstname())
                        .lastname(user.getLastname())
                        .build())
                .toList();
    }

    private UserEntity verifyAccessToken(String authorizationHeader, String requiredScope) {
        if (authorizationHeader != null && authorizationHeader.startsWith("Bearer ")) {
            String accessToken = authorizationHeader.replace("Bearer ", "");

            DecodedJWT decodedJwt;
            Algorithm algorithm = Algorithm.RSA256(jwtPublicKey);
            try {
                decodedJwt = JWT.require(algorithm)
                        .withIssuer("urn:lukasvydrzel")
                        .withAudience("urn:demoapp")
                        .build()
                        .verify(accessToken);
            } catch (Exception e) {
                log.warn("Invalid JWT", e);
                throw new ResponseStatusException(UNAUTHORIZED);
            }

            if (!StringUtils.containsIgnoreCase(decodedJwt.getClaim("scope").asString(), requiredScope)) {
                throw new ResponseStatusException(FORBIDDEN);
            }

            return userDao.findByLoginIgnoreCase(decodedJwt.getSubject())
                    .orElseThrow(() -> new ResponseStatusException(HttpStatus.UNAUTHORIZED));
        }

        throw new ResponseStatusException(HttpStatus.UNAUTHORIZED);
    }

}
