package cz.vydrzel.dpczu.service;

import com.auth0.jwt.JWT;
import com.auth0.jwt.algorithms.Algorithm;
import cz.vydrzel.dpczu.api.idp.model.*;
import cz.vydrzel.dpczu.dao.UserDao;
import cz.vydrzel.dpczu.dao.model.UserEntity;
import dev.samstevens.totp.code.CodeVerifier;
import dev.samstevens.totp.code.HashingAlgorithm;
import dev.samstevens.totp.qr.QrData;
import dev.samstevens.totp.qr.QrGenerator;
import dev.samstevens.totp.secret.SecretGenerator;
import dev.samstevens.totp.util.Utils;
import jakarta.annotation.PostConstruct;
import jakarta.transaction.Transactional;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.collections4.map.PassiveExpiringMap;
import org.apache.commons.lang3.RandomStringUtils;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.springframework.http.HttpStatus;
import org.springframework.security.crypto.argon2.Argon2PasswordEncoder;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;

import java.nio.file.Files;
import java.nio.file.Paths;
import java.security.KeyFactory;
import java.security.Security;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.PKCS8EncodedKeySpec;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.TimeUnit;

@Slf4j
@Service
@Transactional
@RequiredArgsConstructor
public class IdpService {

    public static final String JWT_KEY_ID = "demo-idp-lukasvydrzel";
    private final Argon2PasswordEncoder passwordEncoder = new Argon2PasswordEncoder(16, 32, 1, 65536, 10);

    private final UserDao userDao;

    private final SecretGenerator secretGenerator;

    private final QrGenerator qrGenerator;

    private final CodeVerifier codeVerifier;

    private final Map<UUID, Integer> failedLoginAttempts = new PassiveExpiringMap<>(15, TimeUnit.MINUTES);
    private final Map<UUID, Set<String>> usedTotps = new PassiveExpiringMap<>(3, TimeUnit.MINUTES);

    private RSAPrivateKey jwtPrivateKey;
    private RSAPublicKey jwtPublicKey;


    @PostConstruct
    @SneakyThrows
    public void init() {
        Security.addProvider(new BouncyCastleProvider());

        String privateKeyContent = Files.readString(Paths.get("src/main/resources/certificates/jwt-private-key.pem"));
        privateKeyContent = privateKeyContent
                .replace("-----BEGIN RSA PRIVATE KEY-----", "")
                .replaceAll("\\n", "")
                .replace("-----END RSA PRIVATE KEY-----", "");
        byte[] encodedPrivateKeyContent = Base64.decodeBase64(privateKeyContent);
        KeyFactory keyFactory = KeyFactory.getInstance("RSA");
        PKCS8EncodedKeySpec privateKeySpec = new PKCS8EncodedKeySpec(encodedPrivateKeyContent);
        jwtPrivateKey = (RSAPrivateKey) keyFactory.generatePrivate(privateKeySpec);

        CertificateFactory certificateFactory = CertificateFactory.getInstance("X.509");
        X509Certificate x509Certificate = (X509Certificate) certificateFactory.generateCertificate(getClass().getResourceAsStream("/certificates/jwt-certificate.pem"));
        jwtPublicKey = (RSAPublicKey) x509Certificate.getPublicKey();
    }

    public JwksResponse getJwks() {
        java.util.Base64.Encoder encoder = java.util.Base64.getUrlEncoder().withoutPadding();
        return JwksResponse.builder()
                .key(Jwk.builder()
                        .algorithmFamily("RSA")
                        .algorithm("RS256")
                        .usage("sig")
                        .keyId(JWT_KEY_ID)
                        .rsaModulus(base64UrlUIntEncode(jwtPublicKey.getModulus().toByteArray()))
                        .rsaExponent(base64UrlUIntEncode(jwtPublicKey.getPublicExponent().toByteArray()))
                        .build())
                .build();
    }


    public RegisterUserResponse registerUser(RegisterUserRequest registerUserRequest) {
        String passwordSalt = RandomStringUtils.randomAlphabetic(10);
        String secret = secretGenerator.generate();

        UserEntity userEntity = UserEntity.builder()
                .id(UUID.randomUUID())
                .login(registerUserRequest.getLogin())
                .firstname(registerUserRequest.getFirstname())
                .lastname(registerUserRequest.getLastname())
                .role(registerUserRequest.getRole())
                .secret(secret)
                .passwordSalt(passwordSalt)
                .passwordHash(passwordEncoder.encode(registerUserRequest.getPassword() + passwordSalt))
                .build();

        userEntity = userDao.save(userEntity);

        RegisterUserResponse registerUserResponse = RegisterUserResponse.builder()
                .mfaQrCode(getMfaQrCode(userEntity.getLogin(), userEntity.getSecret()))
                .build();

        return registerUserResponse;
    }

    public LoginResponse openIdLogin(LoginRequest loginRequest) {
        UserEntity userEntity = userDao.findByLoginIgnoreCase(loginRequest.getLogin())
                .orElseThrow(() -> new ResponseStatusException(HttpStatus.UNAUTHORIZED));

        Integer failedLoginAttemptsNumber = failedLoginAttempts.getOrDefault(userEntity.getId(), 0);
        if (failedLoginAttemptsNumber > 3) {
            throw new ResponseStatusException(HttpStatus.UNAUTHORIZED);
        }

        if (!passwordEncoder.matches(loginRequest.getPassword() + userEntity.getPasswordSalt(), userEntity.getPasswordHash())) {
            failedLoginAttempts.put(userEntity.getId(), failedLoginAttemptsNumber + 1);
            throw new ResponseStatusException(HttpStatus.UNAUTHORIZED);
        }

        if (!verifyTotp(loginRequest.getTotp(), userEntity.getSecret())) {
            failedLoginAttempts.put(userEntity.getId(), failedLoginAttemptsNumber + 1);
            throw new ResponseStatusException(HttpStatus.UNAUTHORIZED);
        }

        Set<String> usedUserTotps = usedTotps.getOrDefault(userEntity.getId(), new HashSet<>());
        if (usedUserTotps.contains(loginRequest.getLogin())) {
            throw new ResponseStatusException(HttpStatus.UNAUTHORIZED);
        }
        usedUserTotps.add(loginRequest.getLogin());

        return LoginResponse.builder()
                .idToken(generateIdToken(userEntity))
                .accessToken(generateAccessToken(userEntity))
                .build();
    }

    @SneakyThrows
    private String getMfaQrCode(String login, String secret) {
        QrData data = new QrData.Builder()
                .label(login)
                .secret(secret)
                .issuer("DEMO IDP (Lukáš Vydržel)")
                .algorithm(HashingAlgorithm.SHA256)
                .digits(6)
                .period(60)
                .build();

        return Utils.getDataUriForImage(
                qrGenerator.generate(data),
                qrGenerator.getImageMimeType()
        );
    }

    private boolean verifyTotp(String code, String secret) {
        return codeVerifier.isValidCode(secret, code);
    }

    private String generateIdToken(UserEntity userEntity) {
        Algorithm algorithm = Algorithm.RSA256(jwtPublicKey, jwtPrivateKey);
        return JWT.create()
                .withIssuer("urn:lukasvydrzel")
                .withAudience("urn:demoapp")
                .withSubject(userEntity.getLogin())
                .withClaim("firstname", userEntity.getFirstname())
                .withClaim("lastname", userEntity.getLastname())
                .withIssuedAt(Instant.now())
                .withExpiresAt(Instant.now().plusSeconds(3600))
                .withJWTId(UUID.randomUUID().toString())
                .withKeyId(JWT_KEY_ID)
                .sign(algorithm);
    }

    private String generateAccessToken(UserEntity userEntity) {
        Algorithm algorithm = Algorithm.RSA256(jwtPublicKey, jwtPrivateKey);
        return JWT.create()
                .withIssuer("urn:lukasvydrzel")
                .withClaim("client_id", "demo-app-client")
                .withAudience("urn:demoapp")
                .withClaim("scope", userEntity.getRole().getScopes())
                .withSubject(userEntity.getLogin())
                .withIssuedAt(Instant.now())
                .withExpiresAt(Instant.now().plusSeconds(3600))
                .withJWTId(UUID.randomUUID().toString())
                .withKeyId(JWT_KEY_ID)
                .sign(algorithm);
    }

    public String base64UrlUIntEncode(byte[] toEncode) {
        int modulusLength = toEncode.length;
        int firstNonZeroByteIndex = 0;
        while (toEncode[firstNonZeroByteIndex] == 0) {
            firstNonZeroByteIndex++;
        }
        byte[] trimmed = Arrays.copyOfRange(toEncode, firstNonZeroByteIndex, modulusLength);

        return java.util.Base64.getUrlEncoder().withoutPadding().encodeToString(trimmed);
    }

}
